@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # @Author: Richard J. Mathar <mathar@mpia.de>
-# @Date: 2022-11-04
+# @Date: 2022-11-15
 # @Filename: siderostat.py
 # @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
 
@@ -11,13 +11,21 @@ Python3 class for siderostat field angles using homogeneous coordinates
 
 import math
 import numpy
+import datetime
 import astropy.coordinates
 import astropy.time
 import astropy.units
+import astropy.io.fits
+import astropy.wcs
+# import skymakercam # circular import !?!
+# from skymakercam import SkymakerCam
 
 from lvmtipo.mirror import Mirror
 from lvmtipo.target import Target
 from lvmtipo.kmirror import Kmirror
+from lvmtipo.fiber import Fiber
+from lvmtipo.site import Site
+from lvmtipo.ambient import Ambient
 
 __all__ = ['Siderostat']
 
@@ -29,18 +37,21 @@ class Siderostat():
     def __init__(self, zenang=90.0, azang=0.0, medSign=-1, m1m2dist=240.0):
         """ A siderostat of 2 mirrors
         :param zenang: Zenith angle of the direction of the exit beam (degrees)
-                   in the range 0..180. Default is the design value of the LCO LVMT.
+                   in the range 0..180. Default is the design value of the LVMT.
         :type zenang: float
+
         :param azang: Azimuth angle of the direction of the exit beam (degrees)
                    in the range -180..360 degrees, N=0, E=90.
                    Ought to be zero for the LCO LVMT where the FP is north of the
                    siderostat and 180 for the MPIA test setup where the FP is
-                   south of the siderostat..
+                   south of the siderostat.. The default is the angle for LCO.
         :type azang: float
+
         :param medSign: Sign of the meridian flip design of the mechanics.
                        Must be either +1 or -1. Default is the LCO LVMT design as build (in newer
                        but not the older documentation).
         :type medSign: int
+
         :param m1m2dist: Distance between the centers of M1 and M2 in millimeter.
                        The default value is taken from
                        LVM-0098_Sky Baffle Design of 2022-04-18
@@ -79,6 +90,34 @@ class Siderostat():
         self.box[2] = self.b[1]
         self.boy = numpy.cross(self.b, self.box)
 
+#     def parallact(self, site, target, ambi, time):
+#         """
+#         Compute the parallactic angle for that target at that time.
+#         :return: The parallactic angle in radians
+#         :rtype: float
+#         """
+#         #define the zenith in the topocentric horizontal frame
+#         earthloc = site.toEarthLocation()
+#         zeni_hori = astropy.coordinates.AltAz(
+#             location=earthloc,
+#             obstime=time,
+#             pressure=astropy.units.Quantity(
+#                 100.*ambi.press, unit=astropy.units.Pa),
+#             temperature=astropy.units.Quantity(
+#                 ambi.temp, unit=astropy.units.deg_C),
+#             relative_humidity=ambi.rhum,
+#             obswl=astropy.units.Quantity(ambi.wlen, unit=astropy.units.um),
+#             az=astropy.coordinates.Angle(0,unit=astropy.units.degree),
+#             alt=astropy.coordinates.Angle(90,unit=astropy.units.degree))
+# 
+#         # print(zeni_hori)
+#         # project the zenith to the equatorial coordinates
+#         # zeni= zeni_hori.transform_to(frame='icrs')
+#         icrs_frame = astropy.coordinates.ICRS()
+#         zeni= zeni_hori.transform_to(icrs_frame)
+#         pa = target.targ.position_angle(zeni)
+#         return pa.radian
+
     def fieldAngle(self, site, target, ambi, time=None):
         """
         :param site: geographic ITRF location of the observatory
@@ -154,27 +193,25 @@ class Siderostat():
         # image of targNcp while hitting M1
         m1img = trans.apply(m2tom1)
         # image of targNcp before arriving at M1
-        starOffm1 = m2tom1 + starNcp
-        starimg = trans.apply(starOffm1)
+        star_off_m1 = m2tom1 + starNcp
+        starimg = trans.apply(star_off_m1)
 
         # virtual direction of ray as seen from point after M2
         # no need to normalize this because the atan2 will do...
         starvirt = m1img - starimg
 
         # project in a plane orthogonal to  self.b
-        cosFang = numpy.dot(starvirt, self.box)
-        sinFang = numpy.dot(starvirt, self.boy)
-        return math.atan2(sinFang, cosFang)
+        cos_fang = numpy.dot(starvirt, self.box)
+        sin_fang = numpy.dot(starvirt, self.boy)
+        return math.atan2(sin_fang, cos_fang)
 
     def to_header(self, site, target, camera, ambi, k_mocon_steps, ag_cam_name,
-           genrevx, genrevy, kmirr, time=None,
-           dist_cam_edge=11.14771):
+           genrevx, genrevy, kmirr, time=None,pixsize=None,bin=None,wd=None,hd=None,
+           flen=1839.8, dist_cam_edge=11.14771):
         """ Convert the parameters to FITS WCS header cards.
         This traces the position angle implied by the siderostat and number of mirror
         reflections through the K-mirror if applicable and through the
-        prisms if applicable to predict the key WCS keywords. The main difference
-        to the implementation in lvmcam/python/lvmcam/models/wcs.py
-        is that the variable genvrev[xy] are included in this computation.
+        prisms if applicable to predict the key WCS keywords.
         .. warn:: this assumes that the FITS files are read out as
            in the MPIA test setup: all east/west/center cameras with the
            long edge up-down, no rotation by any 90 or 180 degrees in any
@@ -187,6 +224,7 @@ class Siderostat():
         :type target: astropy.coordinates
 
         :param camera: the scales in the image, pixel and binning etc
+            Can be None, but then the pixsize, bin, wd, hd and flen parameters must be set
         :type camera: SkymakerCamera
 
         :param ambi: Ambient data relevant for refractive index
@@ -199,8 +237,7 @@ class Siderostat():
              various different conventions are in use for this instrument,
              both with respect to sign and with respect to mechanical vs
              optical angles.
-             For one of the three benches (hidden in ag_cam_name) this is
-             a dummy argument because there is no K-mirror.
+             If kmirr=None this parameter is not used.
         :type k_mocon_steps: int
 
         :param ag_cam_name: The telescope (skye, skyw, spec, sci) and camera
@@ -220,7 +257,28 @@ class Siderostat():
         :param time: time of the observation /UTC; if None, the current time will be used.
         :type time:
 
+        :param pixsize: pixel size in microns, not binned
+              Can be None if camera has this information
+        :type pixsize: float
+
+        :param bin: binning factor in the image
+              Can be None if camera has this information
+        :type bin: int
+
+        :param wd: width of the image in pixels (after binning)
+              Can be None if camera has this information
+        :type wd: int
+
+        :param hd: height of the image in pixels (after binning)
+              Can be None if camera has this information
+        :type hd: int
+
+        :param flen: focal lens of lens in mm.
+              Can be None if camera has this information
+        :type flen: float
+
         :param kmirr: K-mirror constants related to conversions of steps to angles
+             May be None to indicate that there is no K-mirror (like for the spec telescope)
         :type kmirr: lvmtipo.Kmirror
 
         :param dist_cam_edge: the distance of the long east or west FLIR camera edge to
@@ -232,11 +290,12 @@ class Siderostat():
         :return a set of FITS header cards with WCS keywords.
         :rtype astropy.io.fits.Header
         """
-        # start with an empty set of header cards
-        wcshdr = astropy.io.fits.Header()
+
+        # start with set of header cards of geographics of the observatory
+        wcshdr = site.to_header()
 
         # compute the position angle (NCP) after the siderostat
-        # and before the K-mirror in radians.
+        # (before the K-mirror) in radians.
         if isinstance(time, astropy.time.Time):
             now = time
         elif isinstance(time, str):
@@ -251,8 +310,14 @@ class Siderostat():
         # Simplify the interface by also recognizing the descr of
         # the camera (which uses upper case for some letter).
         ag_cam_name = ag_cam_name.lower()
+
         # three LVMT benches have a K-mirror, one has not
-        if ag_cam_name.find('skye') >= 0 or ag_cam_name.find('skyw') >= 0 or ag_cam_name.find('sci') >= 0:
+        if kmirr is None:
+            # this will eventually be the case of the spec-telescope
+            # the angle passes unchanged
+            pass
+        else:
+            # this will eventually be the case of the sci, skye, skyw telescopes
             n_refl += 3
 
             # convert K-mirror orientation to radians
@@ -264,10 +329,18 @@ class Siderostat():
             # (ang_in+ang_out)/2 = kangle+-pi/2
             # ang_out = 2*kangle+-pi - ang_in
             ang = math.pi + 2.*kangle - ang
-        elif ag_cam_name.find('spec') >= 0:
-            pass
-        else:
-            raise NameError("Unrecognized camera name " + ag_cam_name)
+
+        tele='LVMT'
+        if ag_cam_name.find('skye') >= 0 :
+           tele +=' skye'
+        elif ag_cam_name.find('skyw') >= 0 :
+            tele +=' skyw'
+        elif ag_cam_name.find('sci') >= 0 :
+           tele +=' sci'
+        elif ag_cam_name.find('spec') >= 0 :
+           tele +=' spec'
+        key = astropy.io.fits.Card("TELESCOP", tele, "Local Volume Mapper sci, skye, skyw, spec")
+        wcshdr.append(key)
 
         if ag_cam_name.find('center') >= 0 or ag_cam_name.find('agc') >= 0:
             # no prism
@@ -293,8 +366,8 @@ class Siderostat():
             ang *= -1.0
             n_refl += 1
             # assume camera is mounted vertically with the long lower edge
-            # pointing north (basically rotated 180 deg around the vertical
-            # axis relative to the west camera) toward the focal plane
+            # pointing south (basically rotated 180 deg around the horizontal
+            # axis relative to the west camera) away from the FP.
             # We want to express an angle where the reference is "up" to an
             # angle where the reference "right" or geographically "to the south"
             # (passive rotation)
@@ -302,11 +375,19 @@ class Siderostat():
         else:
             raise NameError("Unrecognized camera name " + ag_cam_name)
 
-        # assign an image parity defined by the number of mirror reflections
-        if (n_refl % 2) == 0:
-            imgparity = True
+        if ag_cam_name.find('center') >= 0 or ag_cam_name.find('agc') >= 0:
+            tele += '.agc'
+        elif ag_cam_name.find('west') >= 0 or ag_cam_name.find('agw') >= 0:
+            tele += '.agw'
+        elif ag_cam_name.find('east') >= 0 or ag_cam_name.find('age') >= 0:
+            tele += '.age'
         else:
-            imgparity = False
+            tele += '.'
+        key = astropy.io.fits.Card("CAMERA", tele, " AG camera age, agw, agc on the bench")
+        wcshdr.append(key)
+
+        # assign an image parity defined by the even/odd number of mirror reflections
+        imgparity = True if ( (n_refl % 2) == 0 ) else False
 
         # up to here ang is refering to the short-edge-up orientation
         # of the camera if that were read out by streaming the araviscam
@@ -343,26 +424,43 @@ class Siderostat():
         wcshdr.append(key)
         key = astropy.io.fits.Card("CTYPE2", "DEC--TAN", "WCS type axis 2")
         wcshdr.append(key)
+
+        # hexrep = target.targ.to_string('hmsdms')
+        hexrep = target.targ.ra.to_string(unit=astropy.units.hour,precision=2)
         key = astropy.io.fits.Card(
-            "CRVAL1", target.ra.degree, "[deg] RA at reference pixel")
+            "CRVAL1", target.targ.ra.degree, "[deg] RA ref pixel " + hexrep)
         wcshdr.append(key)
+        hexrep = target.targ.dec.to_string(unit=astropy.units.degree,precision=2)
         key = astropy.io.fits.Card(
-            "CRVAL2", target.dec.degree, "[deg] Dec at reference pixel")
+            "CRVAL2", target.targ.dec.degree, "[deg] Dec ref pixel " + hexrep)
         wcshdr.append(key)
 
-        # convert disgtcamEdgeCtr from mm to pixels
-        # this is the distance from the long edge that is narrowest to
+        # convert disgtcamEdgeCtr from mm to micron and then to pixels
+        # This is the distance from the long edge that is narrowest to
         # the fiber center after projection into the focal plane.
-        dist_cam_edge *= 1000.0 / (camera.pixsize*camera.binning[1])
-        # distance from the middle of the east or west camera to fiber ctner in pixels
-        dist_cam_mid = camera.detector_size.hd / \
-            2 / camera.binning[1] + dist_cam_edge
+        if camera is None:
+            dist_cam_edge *= 1000.0 / (pixsize*bin)
+        else:
+            dist_cam_edge *= 1000.0 / (camera.pixsize*camera.binning[1])
 
+        # distance from the middle of the east or west camera to fiber ctner in pixels
+        if camera is None:
+            dist_cam_mid = hd / 2 + dist_cam_edge
+        else:
+            dist_cam_mid = camera.detector_size.hd / \
+                2 / camera.binning[1] + dist_cam_edge
+    
         # where is the center of image away in the pixel coordinate system
-        # of the camera? For the x-position this is in the middel of the
+        # of the camera? For the x-position this is in the middle of the
         # camera along the long axis because all cameras are installed up-down
-        crpix1 = camera.detector_size.wd / 2 / camera.binning[0] + 0.5
-        crpix2 = camera.detector_size.hd / 2 / camera.binning[1] + 0.5
+        # For age and agw the wd parameters are 1600 and the hd 1100 (not binned).
+        if camera is None:
+            crpix1 = wd / 2 + 0.5
+            crpix2 = hd / 2 + 0.5
+        else :
+            crpix1 = camera.detector_size.wd / 2 / camera.binning[0] + 0.5
+            crpix2 = camera.detector_size.hd / 2 / camera.binning[1] + 0.5
+    
         if ag_cam_name.find('center') >= 0 or ag_cam_name.find('agc') >= 0 :
             # ra/dec in the center of the image
             pass
@@ -375,11 +473,11 @@ class Siderostat():
                 crpix2 -= dist_cam_mid
         elif ag_cam_name.find('east') >= 0 or ag_cam_name.find('age') >= 0 :
             # the direction from the prism center to the fiber center
-            # is walkign west, which is in the camera walking to the upper edge
+            # is walkign west, which is in the camera walking to the lower edge
             if genrevy:
-                crpix2 -= dist_cam_mid
-            else:
                 crpix2 += dist_cam_mid
+            else:
+                crpix2 -= dist_cam_mid
 
         key=astropy.io.fits.Card(
             "CRPIX1", crpix1, "[px] point cntr along axis 1")
@@ -389,8 +487,19 @@ class Siderostat():
         wcshdr.append(key)
 
         # the two main values in the CD matrix, irrespective of sign
-        cosperpix = camera.degperpix*math.cos(ang)
-        sinperpix = camera.degperpix*math.sin(ang)
+        if camera is None:
+            # plate scale, microns in the focal plane per radian on the sky
+            # image scale at focus is 8.92 um/arcsec = 8.92um/(pi/180/3600 rad) =1.8e^6 um/rad
+            pscal=1000.*flen
+            # plate scale, pixels in the focal plane per radian on the sky
+            pscal /= pixsize
+            # plate scale, pixels in the focal plane per degrees on the sky
+            pscal = math.radians(pscal)
+            cosperpix = math.cos(ang)/pscal
+            sinperpix = math.sin(ang)/pscal
+        else:
+            cosperpix = camera.degperpix*math.cos(ang)
+            sinperpix = camera.degperpix*math.sin(ang)
         # suppose the parity is positive and ang is small, then  in matrix notation
         # ( x)     ( sin ang, -cos ang ) (delta)
         #       =
@@ -430,8 +539,285 @@ class Siderostat():
             "CD2_2", cosperpix, "[deg/px] WCS matrix diagonal")
         wcshdr.append(key)
 
+        horiz = target.toHoriz(site=site, ambi=ambi, time=now)
+        # print(horiz)
+        # print("alt ", horiz.alt, type(horiz.alt))
+        # print("az ", horiz.az, type(horiz.az))
+        # print("az ", horiz.az.deg)
+        # why does horiz.alt.degrees(), horiz.az.degrees() not work here?
+        # print(horiz.az.radian)
+        # float(horiz.alt.to_string(unit='degree',decimal=True))
+        key=astropy.io.fits.Card(
+            "ALT", horiz.alt.deg, "[deg] object altitude above horizon")
+        wcshdr.append(key)
+        key=astropy.io.fits.Card(
+            "AZ", horiz.az.deg, "[deg] object azimuth N=0, E=90")
+        wcshdr.append(key)
+  
+        # compute the parallactic angle 
+        pa = target.parallact(site, ambi, now)
+        # print(pa, math.degrees(pa))
+        key=astropy.io.fits.Card("PARANG", math.degrees(pa), "[deg] parallactic angle")
+        wcshdr.append(key)
+
+        # update DATE
+        fwritetime = datetime.datetime.utcnow()
+        key=astropy.io.fits.Card(
+            "DATE", fwritetime.strftime("%Y-%m-%dT%H:%M:%S"), "UTC " + str(type(self)) + " keys modification")
+        wcshdr.append(key)
+
         return wcshdr
 
+    def to_wcs(self, site, target, camera, ambi, k_mocon_steps, ag_cam_name,
+           genrevx, genrevy, kmirr, time=None,pixsize=None,bin=None,wd=None,hd=None,flen=1839.8,
+           dist_cam_edge=11.14771):
+        """ Convert the parameters to astropy WCS information.
+        This is a simple convenience wrapper around to_header().
+
+        :param site: geographic ITRF location of the observatory
+        :type site: lvmtipo.Site
+
+        :param target: sidereal target in ra/dec
+        :type target: astropy.coordinates
+
+        :param camera: the scales in the image, pixel and binning etc
+        :type camera: SkymakerCamera
+
+        :param ambi: Ambient data relevant for refractive index
+        .. warn:: The pixel scales in the center camera are wrong as of 2022-11-04
+                  because the configuration yml files do not contain the correct 4.5 um
+        :type ambi: lvmtipo.Ambient
+
+        :param k_mocon_steps: The location of the K-mirror at the time
+             in units of Mocon steps. We avoid the use of bare angles because
+             various different conventions are in use for this instrument,
+             both with respect to sign and with respect to mechanical vs
+             optical angles.
+             If kmirr=None this argument is not used.
+        :type k_mocon_steps: int
+
+        :param ag_cam_name: The telescope (skye, skyw, spec, sci) and camera
+             location (age, agw, agc) or (east, west, center)
+             of the FLIR camera that took the image. The string must be some
+             concatenation of both names.
+        :type ag_cam_name: string
+
+        :param genrevx: True if the readout direction along x was reversed
+             in the araviscam interface before creating the FITS file.
+        :type genrevx: bool
+
+        :param genrevy: True if the readout direction along y was reversed
+             in the araviscam interface before creating the FITS file.
+        :type genrevy: bool
+
+        :param time: time of the observation /UTC; if None, the current time will be used.
+        :type time:
+
+        :param pixsize: pixel size in microns, not binned
+              Can be None if camera has this information
+        :type pixsize: float
+
+        :param bin: binning factor in the image
+              Can be None if camera has this information
+        :type bin: int
+
+        :param wd: width of the image in pixels (after binning)
+              Can be None if camera has this information
+        :type wd: int
+
+        :param hd: height of the image in pixels (after binning)
+              Can be None if camera has this information
+        :type hd: int
+
+        :param flen: focal lens of lens in mm.
+              Can be None if camera has this information
+        :type flen: float
+
+        :param kmirr: K-mirror constants related to conversions of steps to angles
+             May be None to indicate that there is no K-mirror (like for the spec telescope)
+        :type kmirr: lvmtipo.Kmirror
+
+        :param dist_cam_edge: the distance of the long east or west FLIR camera edge to
+             the fiber bundle centre in millimeters
+        :type dist_cam_edge: float
+
+        :return: a set of astropy WCS keywords.
+        :rtype: astropy.wcs.WCS
+        """
+
+        hdr= self.to_header(site, target, camera, ambi, k_mocon_steps, ag_cam_name,
+           genrevx, genrevy, kmirr, time=time,pixsize=pixsize,bin=bin,wd=wd,hd=hd,flen=flen,
+           dist_cam_edge=dist_cam_edge )
+
+        return astropy.wcs.WCS(hdr)
+
+    def update_fits(self, fits_in, fits_out):
+        """
+        Update the WCS header cards of an existing FITS file with
+        computed a-priori parameters  found in the header.
+        The header parameters OBSERVAT, DATE-OBS, GENREV[XY], BINX,
+            KMIRDROT, CAMNAME etc are fed into the computed position angles for
+            the fits_in image (in the primary header), and fits_out
+            is created with the same image but WCS keywords replaced according to the
+            theory of the LVMT mirror trains.
+        This is essentially a checker for existing FITS files so the result can
+          be compared with astrometry.net solutions, and a demonstrator of
+          how first esimators to run astrometry.net  could be found.
+
+        ..warn..: this works only assuming that some standard conventions in the
+            header data (as applying in Nov 2022) are met.
+        :param fits_in: existing fits file name on local disks
+        :type fits_in: str
+
+        :param fits_out: file name of a non-existing FITS file on local disks to be created
+            To prevent creation of the FITS file this can be None, so effectively
+            the returned WCS structure is computed.
+        :type fits_out: str
+
+        :return: a set of astropy WCS keywords computed from the other header keywords.
+        :rtype: astropy.wcs.WCS
+
+        ..todo..: should probably put in the __main__ of the fieldrotation sample
+
+        Usage example: (first in bash rm tst[WCE].fits in the shell, then thisprog.py, then
+          ds9 -mosaic tst[WCE].fits)
+        #!/usr/bin/env python3
+
+        from lvmtipo.siderostat import Siderostat
+        import astropy.io.fits
+
+        fits_in="/home/mathar/lvm.sci.agcam.west_00001112.fits"
+        fits_out="tstW.fits"
+        sid=Siderostat()
+        sid.update_fits(fits_in,fits_out)
+
+        fits_in="/home/mathar/lvm.sci.agcam.east_00001112.fits"
+        fits_out="tstE.fits"
+        sid=Siderostat()
+        sid.update_fits(fits_in,fits_out)
+
+        fits_in="/home/mathar/lvm.spec.agcam.center_00000731.fits"
+        fits_out="tstC.fits"
+        sid=Siderostat()
+        sid.update_fits(fits_in,fits_out)
+        """
+
+        # grab header of PHDU of fits_in
+        hdu = astropy.io.fits.open(fits_in)
+        hdr = hdu[0].header
+        # extract all the wcs-relevant data
+
+        # this is MPIA or LCO
+        observat = hdr['OBSERVAT'].strip()
+        site = Site(name=observat)
+
+        # ra and dec in degrees
+        ra = hdr['RA'] 
+        dec = hdr['DEC'] 
+        # hack 1: correct the absolutely wrong values of 2022-10-26 for Alcyrone
+        # ra=(3+47.0/60.0+29.19/3600.)*15 ;
+        # dec= 24+6/60.0+14.8/3600. ;
+
+        # horizontal and vertical avariscam flips in the image
+        genrevx = hdr['GENREVX'] 
+        genrevy = hdr['GENREVY'] 
+
+        # K-mirror (middle mirror) position angle in degrees
+        # The KMIRDROT uses a convnetion where this is apparently
+        # the mechanical angle with a sign such that -135 indicates the home position.
+        kmang = -hdr['KMIRDROT']
+
+        # DATE-OBS is in TAI time scales (37 seconds off UTC)
+        time = astropy.time.Time(hdr['DATE-OBS'],format='isot',scale='tai')
+
+        # binning factor 1,2,4
+        bin = hdr['BINX']
+        # FLIR camera type (to deduce the pixel size in microns)
+        flirtyp = hdr['CAMTYPE']
+        ag_cam_name = hdr['CAMNAME'].strip()
+
+        # hack to correct incorrect early FITS headers at MPIA
+        # delete the if-block for files after approx Nov 10 2022
+        if 'center' in ag_cam_name :
+            genrevy=True
+        elif 'east' in ag_cam_name :
+            genrevy=True
+        elif 'west' in ag_cam_name :
+            genrevy=True
+
+        # some less important data to build a refractivity model
+        celsius = hdr['BENTEMP']
+        relhum = hdr['BENHUM']
+        pres = hdr['BENPRESS']
+
+        # binned size of the image
+        wd=hdr['NAXIS1']
+        hd=hdr['NAXIS2']
+
+        # try to guess a telescope bench from the file name
+        tele=''
+        if 'sci' in fits_in :
+            tele += 'sci'
+        elif 'skyw' in fits_in :
+            tele += 'skyw'
+        elif 'skye' in fits_in :
+            tele += 'skye'
+        elif 'spec' in fits_in :
+            tele += 'spec'
+
+        # consistency check: if the telescope is not spec this
+        # value needs to be <= 137 which is a rough estimate for the home switch offset
+        # For the spec telescope there is no k-mirror and this angle does not matter
+        if 'spec' not in tele:
+            if abs(kmang) > 200:
+                # for the sci, skye and skyw we need an angle
+                # but this seems not to be correct here...
+                raise ValueError("unsupported K-mirror angle " + str(kmang) + " on " + tele)
+            else:
+                kmirr = Kmirror()
+                k_mocon_steps=kmirr.radians_to_steps(math.radians(kmang))
+        else:
+            kmirr = None
+            k_mocon_steps=None
+
+
+        if relhum >= 0 and pres >0 and celsius > -20 :
+            ambi=Ambient(press=pres, temp=celsius, rhum=relhum)
+        else:
+            ambi=Ambient()
+        obj = astropy.coordinates.SkyCoord(ra=ra, dec=dec,unit="deg")
+        target=Target(obj)
+        # print(target)
+
+        # Because the arcsecs per pixel kw in the FITS files is usually wrong
+        # for the center camera, we derive our own patched values here....
+        if '70S7C' in flirtyp :
+            pixsize=4.5
+        else:
+            pixsize=9.0
+
+        wcshdr = self.to_header(site, target, None, ambi, k_mocon_steps, tele + " " + ag_cam_name, 
+                genrevx, genrevy, kmirr, time, pixsize, bin=bin, wd=wd, hd=hd)
+        # print(wcshdr)
+
+        if fits_out is not None :
+            # In this case we want to create a new not yet existing FITS file
+            # copy the data/image to a new HDU
+            hduout = astropy.io.fits.PrimaryHDU(hdu[0].data)
+    
+            # copy also the header.
+            # Need copy() here, otherwise the update() further down will refuse to replace 
+            # the old CD values which are in the old FITS file.
+            hduout.header = hdu[0].header.copy()
+            hduout.header.update(wcshdr,unique=True)
+            # todo: remove/replace CHECKSUM and edit DATE, currently astropy
+            # will (errnously) not refer to UTC in the checksum/datasum headers
+            # but to the local timezone of the computer
+            hduout.writeto(fits_out,checksum=True)
+
+        wcs= astropy.wcs.WCS(hdr)
+        # print(wcs)
+        return wcs
 
     def centrTarg(self, site, target, ambi, fib, flen=1839.8, time=None):
         """
@@ -449,19 +835,24 @@ class Siderostat():
         the center of the fiber bundle as a distance and position
         angle relativ to the nominal target...
         :param site: location of the observatory
-        :type site: fieldrotation.Site
+        :type site: lvmtipo.Site
+
         :param target: sidereal target in ra/dec
         :type target: astropy.coordinates
+
         :param ambi: Ambient data relevant for refractive index
         :type ambi:
+
         :param fib: One of the fibers in one of the 4 tables
         :type fib: Fiber
+
         :param flen: Focal length (defining th eplate scale) in mm
                     This should be the same as in the YAML file of the cameras...
         :type flen: float
+
         :param time: time of the observation /UTC; if None, the current time will be used.
         :type time:
-        :return a new target which is in the fiber bundle center when target is at the fiber
+        :return: a new target which is in the fiber bundle center when target is at the fiber
         """
 
         # direction to NCP from target in the FP orientation (=0 if NCP=up)
@@ -517,7 +908,7 @@ class Siderostat():
         format [[time0,vel0,pol0,acce0,jer0],[time1,vel1,],[],...]
 
         :param site: location of the observatory
-        :type site: fieldrotation.Site
+        :type site: lvmtipo.Site
 
         :param target: sidereal target in ra/dec
         :type target: astropy.coordinates.SkyCoord
@@ -530,6 +921,8 @@ class Siderostat():
               a fiber selector.
               A value of zero means the +delta is up in the laboratory,
               the value +90 means the +delta direction is right (horizontally E)
+              If this is a fiber, the direction from the center of the fiber bundle
+              to that fiber is implied.
         :type degNCP: float
 
         :param deltaTime: Time covered by a single polynomial in seconds
@@ -608,6 +1001,15 @@ class Siderostat():
           means whether they are in a zenith angle < 60 deg or above the horizon
           at that epoch or date.
         """
+
+        # in case the angle is implicitly coded as a fiber, take the fiber
+        # position as the desired angle in the focal plane. From here on
+        # posang_fp is in radians.
+        if isinstance(degNCP, Fiber) :
+            posang_fp = degNCP.labAngle()
+        else:
+            posang_fp = math.radians(degNCP)
+
         moc=[]
         if polyN > 0:
             if isinstance(time, astropy.time.Time):
@@ -651,29 +1053,23 @@ class Siderostat():
             # plane is supposed to be. Then another 90 deg rotation
             # defines where the K-mirror "up-down" incidence plane should be,
             # which is the (mechanical) plane of the K-mirror motor.
-            rads=[(math.pi + r + math.radians(degNCP))/2. for r in rads]
+            rads=[(math.pi + r + posang_fp)/2. for r in rads]
 
             # Use an optional jump of 180 deg (that's optically 360 deg)
             # to keep trajectory near the angle of stepsAtStart.
-            kmirr = Kmirror(home_is_west = homeIsWest, home_offset = homeOffset, steps_per_turn=stepsPerturn)
-            radsAtStart= kmirr.steps_to_radians( stepsAtStart)
+            kmirr = Kmirror(home_is_west = homeIsWest, home_offset = homeOffset,
+                steps_per_turn=stepsPerturn)
+            radsAtStart = kmirr.steps_to_radians(stepsAtStart)
 
+            # if we are too far away from the desired initial position,
+            # add or subtract 180 degrees .
             if rads[0] < radsAtStart - 0.5*math.pi:
                 rads=[r + math.pi for r in rads]
             elif rads[0] > radsAtStart + 0.5*math.pi:
                 rads=[r - math.pi for r in rads]
 
             # convert all angles from radians to steps
-            rads=[r * stepsPerturn/(2.0*math.pi) for r in rads]
-
-            homeOffsetSteps=homeOffset * stepsPerturn / 360.0
-            # convert all counts measured from 0=up to counts
-            # relative to the home/reference position of the MoCon
-            if homeIsWest:
-                rads=[r + homeOffsetSteps for r in rads]
-            else:
-                rads=[homeOffsetSteps - r for r in rads]
-
+            rads=[ kmirr.radians_to_steps(r) for r in rads]
 
             # 1 cycle = 614.4 microsecs, see section 9.3 of MoCon User's Guide
             cycsteps=deltaTime/614.4e-6
@@ -681,10 +1077,12 @@ class Siderostat():
                 # Scale velocity and acceleration with 2^16=65536, yerk with 2^32.
                 # We do not use acceleration and yerk (almost 0 for LVMT)
                 pos=round(rads[poly])
+
                 # rads[poly+1]-rads[poly]  is velocity in units of counts
                 # per deltaTime. Divide by deltaTime to get counts per second
                 # and multiply by 614.4e-6 to get counts per cycle.
                 vel=round(65536*(rads[poly+1]-rads[poly])/cycsteps)
+
                 # Note the order: duration, velocity is before position....
                 traj=[round(cycsteps), vel, pos, 0, 0]
                 moc.append(traj)
@@ -737,10 +1135,10 @@ class Siderostat():
         '''
 
         # vector T measured from M2 to t with 3 Cartesian coordinates
-        T=numpy.array([t[0]-m2[0], t[1]-m2[1], t[2]-m2[2]], numpy.double)
+        m2_to_t=numpy.array([t[0]-m2[0], t[1]-m2[1], t[2]-m2[2]], numpy.double)
 
         # second auxiliary base vector
-        bCrossT=numpy.cross(self.b, T)
+        bCrossT=numpy.cross(self.b, m2_to_t)
         # length second auxiliary base vector
         lenbCrossT=numpy.linalg.norm(bCrossT)
 
@@ -764,7 +1162,7 @@ class Siderostat():
 
 
         # compute vector s from M1 to target
-        s=numpy.subtract(T, m)
+        s=numpy.subtract(m2_to_t, m)
         slen=numpy.linalg.norm(s)
 
         # normalize s to unit length
@@ -787,7 +1185,7 @@ class CalibScreen():
         Default parameters digitized from Fig 2-1 of
         SDSS-V_0111_LVMi_ICD_Telescope_to_Enclosure_Oct21_Rev-revGB.docx
     """
-    def __init__(self, height=2184, slope=0.29157, tablOffs=460):
+    def __init__(self, height=2184, slope=0.29157, tabl_off=460):
         """ This is a v-shaped geometry that helps to relate
         positions at the ceiling above the 4 telescopes to the M2 mirrors of
         the telescopes.
@@ -802,7 +1200,7 @@ class CalibScreen():
                   of this gives 0.283 rad = 16.25 deg.
         :type slope: float
 
-        :param tablOffs: There is a sort of natural east-west coordinate origin
+        :param tabl_off: There is a sort of natural east-west coordinate origin
                   in the middle between the two M2 of the sci and spec tables
                   (that is 1700/2 = 850 mm away from both). If we use this spot as the
                   zero-value of the x-coordinate, this tablOff is the distance
@@ -810,11 +1208,11 @@ class CalibScreen():
                   The default is that the top is 460 mm away from the origin,
                   which means 850-460mm = 360 mm away from one and 850+460=1310 mm
                   away from the other of the two middle telescopes.
-        :type tablOffs: float
+        :type tabl_off: float
         """
         self.height=height
         self.slope=slope
-        self.eastOffs=tablOffs
+        self.east_off=tabl_off
 
     def cart_coord(self, x_east_pos):
         """
@@ -831,7 +1229,7 @@ class CalibScreen():
         # assume roof is symmetric east to west
         # So this value is zero if the position is at the maximum
         # height of the inner side of the roof.
-        x_rel_roof=abs(x_east_pos-self.eastOffs)
+        x_rel_roof=abs(x_east_pos-self.east_off)
 
         # equation of z coordinate is abscisa intersection with 2.184 m above
         # M2 and slope dz/dy
